#include <exec/lists.h>
#include <exec/memory.h>
#include <dos/dos.h>
#include <dos/dosextens.h>
#include <dos/dostags.h>
#include <proto/dos.h>
#include <proto/exec.h>
#include <stddef.h>
#include <string.h>

#include "status_version.h"

static const char version[] = VERSTAG;


const UBYTE template[]=	"PROCESS/N,"
"FULL/S,"
"TCB/S,"
"CLI=ALL/S,"
"COM=COMMAND/K";

#define ForeachNode(l,n) \
for (n=(void *)(((struct List *)(l))->lh_Head); \
((struct Node *)(n))->ln_Succ; \
n=(void *)(((struct Node *)(n))->ln_Succ))

struct ArgList
{
    ULONG *process;
    ULONG full;
    ULONG tcb;
    ULONG cli;
    STRPTR com;
};


void	PrintProcess(struct DosLibrary	*DOSBase,
                  struct ArgList	*MyArgList,
                  struct MsgPort	*MyMsgPort,
                  ULONG		Num);

int	main(void)
{
	struct DosLibrary	*DOSBase;
	struct Library *SysBase=*((struct Library**) 4);
	struct RDArgs *readarg;
	ULONG Found=FALSE;
	struct ArgList arglist;
	struct ArgList *MyArgList=&arglist;
	int error=RETURN_OK;

	memclr(&arglist,sizeof(arglist));

	if ((DOSBase=(struct DosLibrary *)OpenLibrary("dos.library",37)))
	{
		if ((readarg=ReadArgs(template,(LONG *)&arglist,0)))
		{
#ifdef __MORPHOS__
			if (LIB_MINVER(&DOSBase->dl_lib, 51, 51))
			{
				struct CLIData *cd;

				cd = QueryCLIDataTags(MyArgList->process ? CLIDATATAG_CLINumber : TAG_IGNORE, MyArgList->process ? *MyArgList->process : 0,
				                      MyArgList->com ? CLIDATATAG_CommandName : TAG_IGNORE, (IPTR)MyArgList->com,
				                      CLIDATATAG_Sorted, TRUE,
				                      TAG_DONE);
				if (cd)
				{
					int i;
					for (i = 0; i < cd->cd_NumCLIs; i++)
					{
						const struct CLIDataItem *cdi = cd->cd_CLI[i];

						if (SetSignal(0L, SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C)
						{
							error = RETURN_FAIL;
							SetIoErr(ERROR_BREAK);
							break;
						}

						Found = TRUE;

						if (MyArgList->com)
						{
							Printf("%2ld\n", cdi->cdi_CLINum);
						}
						else
						{
							Printf("Process %2ld: ", cdi->cdi_CLINum);

							if (MyArgList->tcb || MyArgList->full)
							{
								Printf("stk %7ld, gv %3ld, pri %3ld", cdi->cdi_DefaultStack << 2,
								       cdi->cdi_GlobVec, cdi->cdi_Pri);
							}

							if (MyArgList->tcb && !MyArgList->full)
								PutStr("\n");
							else if (cdi->cdi_Flags & 1)
							{
								Printf(" Loaded as command: %s\n", cdi->cdi_Command);
							}
							else
							{
								PutStr(" No Command loaded\n");
							}
						}
					}

					FreeCLIData(cd);
				}
			}
			else
#endif
			{
				struct CliProcList *cpl;
				int last = 1;
				int current = 1;

				ForeachNode(&DOSBase->dl_Root->rn_CliList, cpl)
				{
					struct MsgPort **p = cpl->cpl_Array;
					LONG num = (LONG)*p++;
					LONG k;

					last = cpl->cpl_First + num;

					for (k = 0; k < num; ++k, ++p)
					{
						if(SetSignal(0L,SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C)
						{
							error = RETURN_FAIL;
							SetIoErr(ERROR_BREAK);
							break;
						}

						if (*p)
						{
							if (MyArgList->com)
							{
								const ULONG comlen = strlen(MyArgList->com);
								const struct Process *MyProcess = (struct Process *) (((UBYTE *) *p) - offsetof(struct Process, pr_MsgPort));
								const struct CommandLineInterface *MyCLI = (struct CommandLineInterface *) BADDR(MyProcess->pr_CLI);
								const UBYTE *comname;
								if (MyCLI && (comname = BADDR(MyCLI->cli_CommandName)) &&
								    (comname[0] == comlen) &&
								    (strnicmp(MyArgList->com, comname+1, (size_t)comname[0])==0) &&
								    ((MyArgList->process) ? ((current+k)==*MyArgList->process) : TRUE))
								{
									Printf("%2ld\n", current+k);
									Found = TRUE;
								}
							}
							else if ((!MyArgList->process) ||
							         ((current+k)==*MyArgList->process))
							{
								PrintProcess(DOSBase,
								             &arglist,
								             *p,
								             current+k);
								Found	=	TRUE;
							}
						}
					}
					current+=num;
				}
			}

			if ((!Found) &&
			        (MyArgList->com))
			{
				error = RETURN_WARN;
			}
			else if ((!Found) &&
			         (MyArgList->process))
			{
				Printf("Process %ld does not exist\n",
				       *MyArgList->process);
				error = RETURN_FAIL;
			}

			FreeArgs(readarg);
		}
		else
		{
			//PutStr("wrong args!\n");
			error = RETURN_FAIL;
		}
		if (error != RETURN_OK)
		{
			PrintFault(IoErr(), NULL);
		}
		CloseLibrary((struct Library *)DOSBase);
	}
	return(error);
}

void	PrintProcess(struct DosLibrary	*DOSBase,
                  struct ArgList	*MyArgList,
                  struct MsgPort	*MyMsgPort,
                  ULONG		Num)
{
	struct Process	*MyProcess;
	struct CommandLineInterface	*MyCLI;

	MyProcess = (struct Process*) ((UBYTE*) MyMsgPort - offsetof(struct Process,pr_MsgPort));
	MyCLI	= (struct CommandLineInterface*) BADDR(MyProcess->pr_CLI);

	Printf("Process %2ld: ",
	       Num);

	if (MyCLI)
	{
		if (MyArgList->tcb &&
		        (!MyArgList->full))
		{
			Printf("stk %7ld, gv %3ld, pri %3ld\n",
			       MyCLI->cli_DefaultStack << 2,
			       (MyProcess->pr_GlobVec ? *((ULONG *)MyProcess->pr_GlobVec) : 0),
			       MyProcess->pr_Task.tc_Node.ln_Pri);
			return;
		}
		if (MyArgList->full)
		{
			Printf("stk %7ld, gv %3ld, pri %3ld ",
			       MyCLI->cli_DefaultStack << 2,
			       (MyProcess->pr_GlobVec ? *((ULONG *)MyProcess->pr_GlobVec) : 0),
			       MyProcess->pr_Task.tc_Node.ln_Pri);

		}
		if (MyCLI->cli_Module)
		{
			Printf("Loaded as command: %b\n",
			       MyCLI->cli_CommandName);
		}
		else
		{
			Printf("No Command loaded\n");
		}
	}
	else
	{
		Printf("%s has no CLI\n",
		       (ULONG)MyProcess->pr_Task.tc_Node.ln_Name);
	}
}
